open Kholle_localisation_de_point ;;

let recherche_dichotomique
    ( comp : 'a -> 'b -> ordre )
    ( a : 'a ) ( inf : 'b )
    ( e : 'b ensemble ) : 'b =
  let rec recherche_bornee ( debut : int ) ( fin : int ) : 'b =
    if debut = fin then ( if debut = -1 then inf else e.(debut) )
    else let milieu = ( debut + fin + 1 ) / 2 in
      match comp a e.(milieu) with
      | PLUS_PETIT -> recherche_bornee debut ( milieu - 1 ) (* I *)
      | PLUS_GRAND -> recherche_bornee milieu fin (* II *)
      | EGAUX -> e.(milieu) in (* III *)
  recherche_bornee ( -1 ) ( ( Array.length e ) - 1 ) ;;

let recherche_e_n ( e : point ensemble ) ( x : int ) : point =
  let comp ( x : int ) ( p : point ) : ordre =
    let x', _ = p in ordre_int x x' in
  recherche_dichotomique comp x ( 0, 0 ) e ;;

let recherche_abr
    ( comp : 'a -> 'b -> ordre )
    ( a : 'a ) ( inf : 'b )
    ( abr : 'b abr ) : 'b =
  let rec parcours ( sous_abr : 'b abr ) ( element_inf : 'b ) : 'b =
    match sous_abr with
    | Vide -> element_inf
    | Noeud ( e, abr_g, abr_d ) ->
      match comp a e with
      | PLUS_PETIT -> parcours abr_g element_inf
      | PLUS_GRAND -> parcours abr_d e
      | EGAUX -> e in
  parcours abr inf ;;

let recherche_arn
    ( comp : 'a -> 'b -> ordre )
    ( a : 'a ) ( inf : 'b )
    ( arn : 'b arn ) : 'b =
  let comp_arn ( a : 'a ) ( e : 'b * couleur  ) =
    let ( b, _ ) = e in comp a b in
  let b, _ = recherche_abr comp_arn a ( inf, NOIR ) arn in b ;;

let nombre_aretes ( s : sommet ensemble ) : int =
  let rec parcours ( i : int ) ( nombre : int ) : int =
    if i = Array.length s then nombre
    else parcours ( i + 1 ) ( nombre + ( List.length s.(i).sortantes ) ) in
  parcours 0 0 ;;

type dalle = { bornes : int * int ; segments : arete arn } ;;
type partition = dalle ensemble ;;

(* Paramètre = p : un point du plan *)
(* Paramètre = partition : una partition du plan *)
(* Renvoie l'identifiant de la zone du plan dans laquelle se trouvel le point *)
let localisation_de_point ( point : point ) ( partition : partition ) : int =
  let dalle = recherche_dichotomique
      ( fun point dalle -> ( ordre_lex point dalle.bornes ) )
      point
      { bornes = (0,0) ; segments = Vide }
      partition in
  let segment_limite = recherche_arn ordre_point_arete point
      { segment = ( (0, 0), ( 2 * 524287, 0) ) ;
        zone_droite = 0 ;
        zone_gauche = 0 }
      dalle.segments in
  segment_limite.zone_gauche ;;

(* Paramètres = segment : un segment ( un couple de points ) *)
(* Paramètres = born_inf : abscisse de la ligne horizontale de gauche *)
(* Paramètres = born_sup : abscisse de la ligne horizontale de droite *)
(* Renvoie vrai si le segment traverse la dalle *)
let traverse
    ( segment : segment ) ( borne_inf : int ) ( borne_sup : int )  : bool =
  let ( px, _ ), ( qx, _ ) = segment in
  px <= borne_inf && qx >= borne_sup ;;

(* Paramètres = borne_inf : abscisse de la ligne horizontale de gauche *)
(* Paramètres = borne_sup : abscisse de la ligne horizontale de droite *)
(* Paramètres = sommets : l'ensemble des sommets et arêtes dans lequel
   la dalle est construite *)
(* Renvoie la dalle du plan délimitée par borne_inf et borne_sup *)
let creation_dalle
    ( borne_inf : int ) ( borne_sup : int )
    ( sommets : sommet ensemble ) : dalle =
  (* Paramètres = isommet : l'indice du sommet dont on ajoute les arêtes
  sortantes *)
  (* Paramètres = arete_a_ajouter : les aretes sortantes restant à ajouter *)
  (* Paramètres = arete_arn : l'arbre bicolore contenant toutes les arêtes
  traversant la dalle que l'on a rencontré à cette étape du parcours. *)
  (* Renvoie l'arbre bicolore contenant toutes les arêtes traversant la dalle
  délimitée par borne_inf et borne_sup *)
  let rec ajout_aretes
      ( isommet : int )
      ( arete_a_ajouter : arete list )
      ( arete_arn : arete arn ) : arete arn =
    match arete_a_ajouter with
    (* Tous les sommets ont été parcouru *)
    | [] when isommet = ( Array.length sommets ) - 1 -> arete_arn
    (* Toutes les arêtes sortantes du sommet ont été parcourues *)
    | [] -> ajout_aretes ( isommet + 1 ) sommets.( isommet + 1).sortantes arete_arn
    (* On croise une arete qui traverse la dalle *)
    | arete :: liste_arete when traverse arete.segment borne_inf borne_sup ->
      ajout_aretes isommet liste_arete
        ( insertion_arn ordre_lex_aretes arete arete_arn )
    (* On croise une arete qui ne traverse pas la dalle *)
    | _ :: liste_arete -> ajout_aretes isommet liste_arete arete_arn in
  { bornes = ( borne_inf, borne_sup ) ;
    segments = ajout_aretes (-1) [] Vide } ;;

(* Paramètres = sommets : l'ensemble des sommets et arêtes dans lequel
   la dalle est construite. L'ensemble doit contenir au moins un élément. *)
(* Renvoie la partition du plan *)
let creation_partition ( sommets : sommet ensemble )
  : partition =
  (* Paramètres = isommet : l'indice du sommet examiné *)
  (* Paramètres = borne_inf : l'abscisse de la ligne horizontale délimitant la
     gauche de la dalle entrain d'être calculée *)
  (* Paramètres = liste_dalles : la liste des dalles (triée de gauche à droite
     sur le plan) déjà calculées à cette étape du parcours. *)
  (* Renvoie la partition du plan *)
  let rec parcours_dalle
      ( isommet : int )
      ( borne_inf : int )
      ( liste_dalles : dalle list )
    : partition =
    if isommet = ( Array.length sommets )
    then Array.of_list ( List.rev liste_dalles )
    else begin
      let x, _ = sommets.( isommet ).point in
      if x != borne_inf then
        let dalle = ( creation_dalle borne_inf x sommets ) in
        parcours_dalle ( isommet + 1 ) x ( dalle  :: liste_dalles )
      else parcours_dalle ( isommet + 1 ) borne_inf liste_dalles
    end in
  let borne_inf, _ =  sommets.(0).point in
  parcours_dalle 1 borne_inf [] ;;

(* Paramètre = aretes_arn : arbre bicolore contenant des aretes *)
(* Paramètre = arete : la liste des arêtes à supprimer de aretes_arn *)
(* Renvoie un arbre bicolore où toutes la liste des arêtes passées en
paramètre a été supprimée de aretes_arn *)
let rec suppression_entrantes
    ( aretes_arn : arete arn ) ( aretes : arete list ) : arete arn =
match aretes with
| [] -> aretes_arn
| arete :: liste_arete ->
    suppression_entrantes
    ( suppression_arn ordre_lex_aretes arete aretes_arn ) liste_arete ;;

(* Paramètre = aretes_arn : arbre bicolore contenant des aretes *)
(* Paramètre = arete : la liste des arêtes à insérer dans aretes_arn *)
(* Renvoie un arbre bicolore où toutes la liste des arêtes passées en
    paramètre a été insérée dans aretes_arn *)
let rec insertion_sortantes
    ( aretes_arn : arete arn ) ( aretes : arete list ) : arete arn =
  match aretes with
  | [] -> aretes_arn
  | arete :: liste_arete ->
    insertion_sortantes
      ( insertion_arn ordre_lex_aretes arete aretes_arn ) liste_arete ;;

(* Paramètre = zones : l'ensemble des sommets et arêtes partitionnant le plan *)
(* Paramètre = isommet : l'indice du plus petit sommet inclu dans la dalle à
   construire *)
(* Paramètre = segments_dalle_precedente : l'arbre bicolore contenant les
   segments traversant la dalle précédente *)
(* Renvoie la dalle contenant isommet_base et l'indice du plus petit sommet de
   zones appartenant à la dalle suivante *)
let creation_dalle_persistent
    ( zones : sommet ensemble )
    ( isommet_base : int )
    ( segments_dalle_precedente : arete arn )
  : dalle * int =
  let x_inf, _ = zones.(isommet_base).point in
  (* Paramètre = isommet : l'indice du sommet en cours de traitement *)
  (* Paramètre = arete_arn : l'arbre bicolore contenant les segments traversant
     la dalle après traitement des sommets avant isommet *)
  (* Renvoie la dalle contenant isommet_base et l'indice du plus petit sommet
     de zones appartenant à la dalle suivante *)
  let rec traitement_sommet
      ( isommet : int ) ( arete_arn : arete arn ) : dalle * int =
    (* Cas de la dernière dalle, vers l'infini *)
    if isommet = ( Array.length zones )
    then ( { bornes = ( 0, 0 ) ; segments = Vide }, isommet )
    else let x, _ = zones.(isommet).point in
      (* Cas du traitement d'un sommet qui n'est plus dans la dalle de base *)
      if x != x_inf
      then ( { bornes = ( x_inf, x ) ; segments = arete_arn }, isommet )
      else traitement_sommet ( isommet + 1 )
          ( insertion_sortantes
              ( suppression_entrantes arete_arn zones.(isommet).entrantes  )
              zones.(isommet).sortantes ) in
  traitement_sommet isommet_base segments_dalle_precedente ;;

(* Paramètre = zone : l'ensembles des sommets et arêtes partitionnant le plan *)
(* Renvoie une partition du plan *)
let creation_partition_persistent ( zones : sommet ensemble )
  : partition =
  let rec parcours_dalle
      ( isommet : int )
      ( liste_dalles : dalle list )
    : partition =
    let segments_dalle_precedente = match liste_dalles with
      | [] ->  Vide
      | dalle :: liste_dalle -> dalle.segments in
    let dalle, isommet_suivant =
      creation_dalle_persistent zones isommet segments_dalle_precedente in
    (* Fin du parcours des sommets *)
    if isommet_suivant = ( Array.length zones )
    then Array.of_list ( List.rev liste_dalles )
    (* Création de la dalle suivante *)
    else parcours_dalle isommet_suivant  ( dalle :: liste_dalles ) in
  parcours_dalle 0 [] ;;

type 'a gros_arn = Vide | Noeud of 'a * ( ( int * ( 'a gros_arn ) ) arn ) * ( ( int * ( 'a gros_arn ) ) arn ) ;;
type partition_ST = ( int * ( arete gros_arn ) arn )

let recherche_gros_arn
    ( comp : 'a -> 'b -> ordre )
    ( a : 'a ) ( i : int )
    ( inf : 'b )
    ( gros_arn : 'b gros_arn ) : 'b =
  let rec parcours ( gros_arn : 'b gros_arn ) ( element_inf : 'b ) : 'b =
    match gros_arn with
    | Vide -> element_inf
    | Noeud ( e, arn_g, arn_d ) ->
      match comp a e with
      | PLUS_PETIT ->
        let _, fils_gauche =
          recherche_arn ( fun i (j,_) -> ordre_int i j ) i ( 0, Vide ) arn_g in
        parcours fils_gauche element_inf
      | PLUS_GRAND ->
        let _, fils_droit =
          recherche_arn ( fun i (j,_) -> ordre_int i j ) i ( 0, Vide ) arn_d in
        parcours fils_droit element_inf
      | EGAUX -> e in
  parcours gros_arn inf ;;
